# coding: utf-8

import trio
import queue
import signal
import asyncio
import outcome
import greenlet
import contextlib

from ._base import BaseTrioEventLoop, TrioAsyncioExit
from ._loop import current_loop


async def _sync(proc, *args):
    return proc(*args)


# Context manager to ensure all between-greenlet switches occur with
# an empty trio run context and unset signal wakeup fd. That way, each
# greenlet can have its own private Trio run.
@contextlib.contextmanager
def clean_trio_state():
    trio_globals = trio._core._run.GLOBAL_RUN_CONTEXT.__dict__
    old_state = trio_globals.copy()
    old_wakeup_fd = None
    try:
        old_wakeup_fd = signal.set_wakeup_fd(-1)
    except ValueError:
        pass  # probably we're on the non-main thread
    trio_globals.clear()
    try:
        yield
    finally:
        if old_wakeup_fd is not None:
            signal.set_wakeup_fd(old_wakeup_fd, warn_on_full_buffer=(not old_state))
        trio_globals.clear()
        trio_globals.update(old_state)


class SyncTrioEventLoop(BaseTrioEventLoop):
    """
    This is the "compatibility mode" implementation of the Trio/asyncio
    event loop. It runs synchronously, by delegating the Trio event loop to
    a separate thread.

    For best results, you should use the asynchronous
    :class:`trio_asyncio.TrioEventLoop` – if possible.
    """

    _loop_running = False
    _stop_pending = False
    _glet = None

    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)

        # We must start the Trio loop immediately so that self.time() works
        self._glet = greenlet.greenlet(trio.run)
        with clean_trio_state():
            if not self._glet.switch(self.__trio_main):
                raise RuntimeError("Loop could not be started")

    def stop(self):
        """Halt the main loop.

        Any callbacks queued before this point are processed before
        stopping.

        """

        def do_stop():
            self._stop_pending = False
            raise TrioAsyncioExit("stopping trio-asyncio loop")

        if self._loop_running and not self._stop_pending:
            self._stop_pending = True
            self._queue_handle(asyncio.Handle(do_stop, (), self))

    def _queue_handle(self, handle):
        self._check_closed()
        if self._glet is not greenlet.getcurrent() and self._token is not None:
            self.__run_in_greenlet(_sync, self._q_send.send_nowait, handle)
        else:
            self._q_send.send_nowait(handle)
        return handle

    def run_forever(self):
        self.__run_in_greenlet(self._main_loop)

    def is_running(self):
        if self._closed:
            return False
        return self._loop_running

    def _add_reader(self, fd, callback, *args):
        if self._glet is not greenlet.getcurrent() and self._token is not None:
            self.__run_in_greenlet(_sync, super()._add_reader, fd, callback, *args)
        else:
            super()._add_reader(fd, callback, *args)

    def _add_writer(self, fd, callback, *args):
        if self._glet is not greenlet.getcurrent() and self._token is not None:
            self.__run_in_greenlet(_sync, super()._add_writer, fd, callback, *args)
        else:
            super()._add_writer(fd, callback, *args)

    def _trio_io_cancel(self, cancel_scope):
        if self._glet is not greenlet.getcurrent() and self._token is not None:
            self.__run_in_greenlet(_sync, cancel_scope.cancel)
        else:
            cancel_scope.cancel()

    def run_until_complete(self, future):
        """Run until the Future is done.

        If the argument is a coroutine, it is wrapped in a Task.

        WARNING: It would be disastrous to call run_until_complete()
        with the same coroutine twice -- it would wrap it in two
        different Tasks and that can't be good.

        Return the Future's result, or raise its exception.
        """
        return self.__run_in_greenlet(self._run_coroutine, future)

    async def _run_coroutine(self, future):
        """Helper for run_until_complete().

        We need to make sure that a RuntimeError is raised
        if the loop is stopped before the future completes.

        This code runs in the Trio greenlet.
        """
        result = None
        future = asyncio.ensure_future(future, loop=self)

        def is_done(_):
            nonlocal result

            result = outcome.capture(future.result)
            if isinstance(result, outcome.Error) and isinstance(
                result.error, (SystemExit, KeyboardInterrupt)
            ):
                # These exceptions propagate out of the event loop;
                # don't stop the event loop again, or else it will
                # interfere with cleanup actions like
                # run_until_complete(shutdown_asyncgens)
                return
            self.stop()

        future.add_done_callback(is_done)
        try:
            await self._main_loop()
            try:
                while result is None:
                    try:
                        await self._main_loop_one(no_wait=True)
                    except TrioAsyncioExit:
                        pass
            except trio.WouldBlock:
                pass
        finally:
            future.remove_done_callback(is_done)

        if result is None:
            raise RuntimeError("Event loop stopped before Future completed.")
        return result.unwrap()

    def __run_in_greenlet(self, async_fn, *args):
        self._check_closed()
        if self._loop_running:
            raise RuntimeError(
                "You can't nest calls to run_until_complete()/run_forever()."
            )
        if asyncio._get_running_loop() is not None:
            raise RuntimeError(
                "Cannot run the event loop while another loop is running"
            )
        if not self._glet:
            if async_fn is _sync:
                # Allow for cleanups during close()
                sync_fn, *args = args
                return sync_fn(*args)
            raise RuntimeError("The Trio greenlet is not running")
        with clean_trio_state():
            res = self._glet.switch((greenlet.getcurrent(), async_fn, args))
        if res is None:
            raise RuntimeError("Loop has died / terminated")
        return res.unwrap()

    async def __trio_main(self):
        from ._loop import _sync_loop_task_name

        trio.lowlevel.current_task().name = _sync_loop_task_name

        # The non-context-manager equivalent of open_loop()
        async with trio.open_nursery() as nursery:
            await self._main_loop_init(nursery)
            with clean_trio_state():
                req = greenlet.getcurrent().parent.switch(True)

            while not self._closed:
                if req is None:
                    break
                caller, async_fn, args = req

                self._loop_running = True
                asyncio._set_running_loop(self)
                current_loop.set(self)
                result = await outcome.acapture(async_fn, *args)
                asyncio._set_running_loop(None)
                current_loop.set(None)
                self._loop_running = False

                if isinstance(result, outcome.Error) and isinstance(
                    result.error, trio.Cancelled
                ):
                    res = RuntimeError("Main loop cancelled")
                    res.__cause__ = result.error.__cause__
                    result = outcome.Error(res)

                with clean_trio_state():
                    req = caller.switch(result)

            with trio.CancelScope(shield=True):
                await self._main_loop_exit()
            nursery.cancel_scope.cancel()

    def __enter__(self):
        return self

    def __exit__(self, *tb):
        self.stop()
        self.close()
        assert self._glet is None

    def _close(self):
        """Hook to terminate the thread"""
        if self._glet is not None:
            if self._glet is greenlet.getcurrent():
                raise RuntimeError("You can't close a sync loop from the inside")
            # The parent will generally already be this greenlet, but might
            # not be in nested-loop cases.
            self._glet.parent = greenlet.getcurrent()
            with clean_trio_state():
                self._glet.switch(None)
            assert self._glet.dead
            self._glet = None
        super()._close()



import os
import sys
import math
import trio
import heapq
import signal
import asyncio
import warnings
import concurrent.futures

from ._handles import ScopedHandle, AsyncHandle
from ._util import run_aio_future

from selectors import _BaseSelectorImpl, EVENT_READ, EVENT_WRITE
from sniffio import thread_local as sniffio_library

try:
    from trio.lowlevel import wait_for_child
except ImportError:
    from ._child import wait_for_child

import logging

logger = logging.getLogger(__name__)

_mswindows = sys.platform == "win32"

try:
    _wait_readable = trio.lowlevel.wait_readable
    _wait_writable = trio.lowlevel.wait_writable
except AttributeError:
    _wait_readable = trio.lowlevel.wait_socket_readable
    _wait_writable = trio.lowlevel.wait_socket_writable


class _Clear:
    def clear(self):
        pass


# Exception raised internally to stop the main loop. Must subclass
# SystemExit or KeyboardInterrupt in order to make it through various
# asyncio layers.
class TrioAsyncioExit(SystemExit):
    pass


class _TrioSelector(_BaseSelectorImpl):
    """A selector that hooks into a ``TrioEventLoop``.

    In fact it's just a basic selector that disables the actual
    ``select()`` method, as that is controlled by the event loop.
    """

    def select(self, timeout=None):  # pragma: no cover
        raise NotImplementedError

    def _select(self, r, w, x, timeout=None):  # pragma: no cover
        raise NotImplementedError


class TrioExecutor(concurrent.futures.ThreadPoolExecutor):
    """An executor that runs its job in a Trio worker thread.

    Bases: `concurrent.futures.ThreadPoolExecutor`

    Arguments:
      limiter (trio.CapacityLimiter or None): If specified, use this
          capacity limiter to control the number of threads in which
          this exeuctor can be running jobs.
      thread_name_prefix: unused
      max_workers (int or None): If specified and *limiter* is not specified,
          create a new `trio.CapacityLimiter` with this value as its limit,
          and use that as the *limiter*.

    """

    def __init__(self, limiter=None, thread_name_prefix=None, max_workers=None):
        self._running = True
        if limiter is None and max_workers is not None:
            limiter = trio.CapacityLimiter(max_workers)
        self._limiter = limiter

    async def submit(self, func, *args):
        if not self._running:  # pragma: no cover
            raise RuntimeError("Executor is down")
        # there is no need to call self._limiter.acquire() here, run_sync does it
        return await trio.to_thread.run_sync(func, *args, limiter=self._limiter)

    def shutdown(self, wait=None):
        self._running = False


class AsyncGeneratorDispatcher:
    """Helper object providing async generator hooks that route
    finalization to either the correct trio-asyncio event loop or the
    outer Trio run, depending on where the generator was first iterated.
    """

    def __init__(self, prev_hooks):
        self.prev_hooks = prev_hooks
        self.refcnt = 1

    @classmethod
    def install(cls):
        current_hooks = sys.get_asyncgen_hooks()

        # These hooks should either be our own AsyncGeneratorDispatcher
        # (for another trio-asyncio loop) or Trio's hooks. Both of those
        # provide both hooks.
        assert current_hooks.firstiter is not None
        assert current_hooks.finalizer is not None

        matches = (
            getattr(current_hooks.firstiter, "__func__", None) is cls.firstiter
        ) + (getattr(current_hooks.finalizer, "__func__", None) is cls.finalizer)
        if matches == 0:
            # Create a new dispatcher that forwards non-trio-asyncio asyncgens
            # to the current_hooks
            dispatcher = cls(prev_hooks=current_hooks)
            sys.set_asyncgen_hooks(
                firstiter=dispatcher.firstiter, finalizer=dispatcher.finalizer
            )
        else:
            # Take a new reference to the dispatcher that the current_hooks
            # refer to
            assert matches == 2
            dispatcher = current_hooks.firstiter.__self__
            assert dispatcher is current_hooks.finalizer.__self__
            assert isinstance(dispatcher, cls)
            dispatcher.refcnt += 1
        return dispatcher

    def uninstall(self):
        self.refcnt -= 1
        if self.refcnt <= 0:
            sys.set_asyncgen_hooks(*self.prev_hooks)
            assert self.refcnt == 0

    def firstiter(self, agen):
        if sniffio_library.name == "asyncio":
            loop = asyncio.get_running_loop()
            agen.ag_frame.f_locals["@trio_asyncio_loop"] = loop
            return loop._asyncgen_firstiter_hook(agen)
        else:
            return self.prev_hooks.firstiter(agen)

    def finalizer(self, agen):
        try:
            loop = agen.ag_frame.f_locals.get("@trio_asyncio_loop")
        except AttributeError:  # pragma: no cover
            loop = None
        if loop is not None:
            return loop._asyncgen_finalizer_hook(agen)
        else:
            return self.prev_hooks.finalizer(agen)


class BaseTrioEventLoop(asyncio.SelectorEventLoop):
    """An asyncio event loop that runs on top of Trio.

    Bases: `asyncio.SelectorEventLoop`

    All event loops created by trio-asyncio are of a type derived from
    `BaseTrioEventLoop`.

    Arguments:
        queue_len:
            The maximum length of the internal event queue.
            The default of None means unlimited. A limit should be specified
            only if you would rather crash your program than use too
            much memory, because it's not feasible to enforce graceful
            backpressure here.
    """

    # This code implements a semi-efficient way to run asyncio code within Trio.
    # A loop may be in one of four states.
    #
    # * new - sync loops are auto-started when first used
    # * stopped - data structures are live but the main loop is not running
    # * running - events are processed
    # * closed - nothing further may happen

    # for calls from other threads or contexts
    _token = None

    # an event; set while the loop is not running
    # To wait until the loop _is_ running, call ``await loop.synchronize()``.
    _stopped = None

    # asyncio's flag whether the loop has been closed
    _closed = False

    # All sub-tasks are started in here
    _nursery = None

    # (Trio) task this loop is running in
    _task = None

    # (threading) Thread this loop is running in
    _thread = None

    # An instance of AsyncGeneratorDispatcher for handling asyncio async
    # generators; it may be shared by multiple running trio-asyncio loops
    _asyncgen_dispatcher = None

    def __init__(self, queue_len=None):
        if queue_len is None:
            queue_len = math.inf

        # Processing queue
        self._q_send, self._q_recv = trio.open_memory_channel(queue_len)

        # which files to close?
        self._close_files = set()

        # set up
        super().__init__(_TrioSelector())

        # replaced internal data
        self._ready = _Clear()
        self._scheduled = _Clear()
        self._default_executor = TrioExecutor()

        self._orig_signals = {}

        # we do our own timeout handling
        self._timers = []

        # Marker whether the loop is actually running
        self._stopped = trio.Event()
        self._stopped.set()

        self._stop_wait = None

    def __repr__(self):
        try:
            return "<%s running=%s at 0x%x>" % (
                self.__class__.__name__,
                "closed" if self._closed else "no" if self._stopped.is_set() else "yes",
                id(self),
            )
        except Exception as exc:
            return "<%s ?:%s>" % (self.__class__.__name__, repr(exc))

    def set_default_executor(self, executor):
        if not isinstance(executor, TrioExecutor):
            if "pytest" not in sys.modules:
                warnings.warn(
                    "executor %r is not supported by trio-asyncio and will "
                    "not be used" % (repr(executor),),
                    RuntimeWarning,
                    stacklevel=2,
                )
            return
        super().set_default_executor(executor)

    def time(self):
        """Return Trio's idea of the current time."""
        if self._task is None:
            raise RuntimeError("This loop is closed.")
        return self._task._runner.clock.current_time()

    # A future doesn't require a trio_asyncio loop
    run_aio_future = staticmethod(run_aio_future)

    # A coroutine (usually) does.
    async def run_aio_coroutine(self, coro):
        """Schedule an asyncio-flavored coroutine for execution on this loop
        by wrapping it in an `asyncio.Task`. Wait for it to complete,
        then return or raise its result.

        Cancelling the current Trio scope will cancel the coroutine,
        which will throw a single `asyncio.CancelledError` into the coroutine
        (just like the usual asyncio behavior). If the coroutine then
        exits with a `~asyncio.CancelledError` exception, the call to
        :meth:`run_aio_coroutine` will raise `trio.Cancelled`.
        But if it exits with `~asyncio.CancelledError` when the current
        Trio scope was *not* cancelled, the `~asyncio.CancelledError` will
        be passed along unchanged.

        This is a Trio-flavored async function.

        """
        try:
            self._check_closed()
            fut = asyncio.ensure_future(coro, loop=self)
        except BaseException:
            coro.close()  # avoid unawaited coroutine error
            raise
        return await run_aio_future(fut)

    def trio_as_future(self, proc, *args):
        """Start a new Trio task to run ``await proc(*args)`` asynchronously.
        Return an `asyncio.Future` that will resolve to the value or exception
        produced by that call.

        Errors raised by the Trio call will only be used to resolve
        the returned Future; they won't be propagated in any other
        way. Thus, if you want to notice exceptions, you had better
        not lose track of the returned Future.  The easiest way to do this is
        to immediately await it in an asyncio-flavored function:
        ``await loop.trio_as_future(trio_func, *args)``.

        Note that it's the awaiting of the returned future, not the
        call to :meth:`trio_as_future` itself, that's
        asyncio-flavored. You can call :meth:`trio_as_future` in a
        Trio-flavored function or even a synchronous context, as long
        as you plan to do something with the returned Future other
        than immediately awaiting it.

        Cancelling the future will cancel the Trio task running your
        function, or prevent it from starting if that is still possible.
        If the Trio task exits due to this cancellation, the future
        will resolve to an `asyncio.CancelledError`.

        Arguments:
          proc: a Trio-flavored async function
          args: arguments for *proc*

        Returns:
          an `asyncio.Future` which will resolve to the result of the call to *proc*
        """
        f = asyncio.Future(loop=self)
        self._queue_handle(AsyncHandle(proc, args, self, result_future=f))
        return f

    def run_trio_task(self, proc, *args):
        """Start a new Trio task to run ``await proc(*args)`` asynchronously.
        If it raises an exception, allow the exception to propagate out of
        the trio-asyncio event loop (thus terminating it).

        Arguments:
          proc: a Trio-flavored async function
          args: arguments for *proc*

        Returns:
          an `asyncio.Handle` which can be used to cancel the background task
        """
        return self._queue_handle(AsyncHandle(proc, args, self))

    # Callback handling #

    def _queue_handle(self, handle):
        """Queue a :class:`Handle` or :class:`TimerHandle` to be executed
        by the event loop.

        :param handle: A handle for the code to be executed
        :type handle: :class:`Handle` or :class:`TimerHandle`
        :return: the handle, for convenience.
        """
        raise RuntimeError("override me")

    def _call_soon(self, *arks, **kwargs):
        raise RuntimeError("_call_soon() should not have been called")

    def call_later(self, delay, callback, *args, **context):
        """asyncio's timer-based delay

        Note that the callback is a sync function.

        :param delay: Time to wait, in seconds.
        :param callback: Sync function to call.
        :return: a handle which may be used to cancel the timer.
        """
        self._check_callback(callback, "call_later")
        assert delay >= 0, delay
        h = asyncio.TimerHandle(delay + self.time(), callback, args, self, **context)
        self._queue_handle(h)
        return h

    def call_at(self, when, callback, *args, **context):
        """asyncio's time-based delay

        Note that the callback is a sync function.
        """
        self._check_callback(callback, "call_at")
        return self._queue_handle(
            asyncio.TimerHandle(when, callback, args, self, **context)
        )

    def call_soon(self, callback, *args, **context):
        """asyncio's defer-to-mainloop callback executor.

        Note that the callback is a sync function.
        """
        self._check_callback(callback, "call_soon")
        return self._queue_handle(asyncio.Handle(callback, args, self, **context))

    def call_soon_threadsafe(self, callback, *args, **context):
        """asyncio's thread-safe defer-to-mainloop

        Note that the callback is a sync function.
        """
        self._check_callback(callback, "call_soon_threadsafe")
        self._check_closed()
        h = asyncio.Handle(callback, args, self, **context)
        self._token.run_sync_soon(self._q_send.send_nowait, h)

    # drop all timers

    def _cancel_timers(self):
        for tm in self._timers:
            tm.cancel()
        self._timers.clear()

    # supersede some built-ins which should not be used

    def _add_callback(self, handle, _via_token=False):
        raise RuntimeError("_add_callback() should have been superseded")

    def _add_callback_signalsafe(self, handle):  # pragma: no cover
        raise RuntimeError("_add_callback_signalsafe() should have been superseded")

    def _handle_signal(self, signum):
        raise RuntimeError("_handle_signal() should have been superseded")

    def _timer_handle_cancelled(self, handle):
        pass

    # Subprocess handling #

    if not _mswindows:

        async def _make_subprocess_transport(
            self,
            protocol,
            args,
            shell,
            stdin,
            stdout,
            stderr,
            bufsize,
            extra=None,
            **kwargs
        ):
            """Make a subprocess transport. Asyncio context."""

            waiter = self.create_future()
            transp = asyncio.unix_events._UnixSubprocessTransport(
                self,
                protocol,
                args,
                shell,
                stdin,
                stdout,
                stderr,
                bufsize,
                waiter=waiter,
                extra=extra,
                **kwargs
            )

            async def child_wait(transp):
                returncode = await wait_for_child(transp.get_pid())
                transp._process_exited(returncode)

            self.run_trio_task(child_wait, transp)
            try:
                await waiter
            except BaseException as exc:
                # Workaround CPython bug #23353: using yield/yield-from in an
                # except block of a generator doesn't clear properly
                # sys.exc_info()
                err = exc
            else:
                err = None

            if err is not None:
                transp.close()
                await transp._wait()
                raise err

            return transp

    # Thread handling #

    def run_in_executor(self, executor, func, *args):
        """
        Delegate running a synchronous function to a separate thread.

        Limitation:
        The executor must be None, or a (subclass of) ``TrioExecutor``.

        Returns an asyncio.Future.
        """
        self._check_callback(func, "run_in_executor")
        self._check_closed()
        if executor is None:  # pragma: no branch
            executor = self._default_executor
        assert isinstance(executor, TrioExecutor)
        return self.trio_as_future(executor.submit, func, *args)

    async def synchronize(self):
        """Suspend execution until all callbacks previously scheduled using
        ``call_soon()`` have been processed.

        This is a Trio-flavored async function.

        From asyncio, call ``await trio_as_aio(loop.synchronize)()``
        instead of ``await asyncio.sleep(0)`` if you need to process all
        queued callbacks.

        """
        w = trio.Event()
        self._queue_handle(asyncio.Handle(w.set, (), self))
        await w.wait()

    # Signal handling #

    def _handle_sig(self, sig, _):
        """Helper to safely enqueue a signal handler."""
        h = self._signal_handlers[sig]
        self._token.run_sync_soon(self._q_send.send_nowait, h)

    def add_signal_handler(self, sig, callback, *args):
        """asyncio's method to add a signal handler."""
        self._check_closed()
        self._check_signal(sig)
        if sig == signal.SIGKILL:
            raise RuntimeError("SIGKILL cannot be caught")
        h = asyncio.Handle(callback, args, self)
        assert sig not in self._signal_handlers, "Signal %d is already being caught" % (
            sig,
        )
        self._orig_signals[sig] = signal.signal(sig, self._handle_sig)
        self._signal_handlers[sig] = h

    def remove_signal_handler(self, sig):
        """asyncio's method to remove a signal handler."""
        # self._check_signal(sig)
        try:
            h = self._signal_handlers.pop(sig)
        except KeyError:
            return False
        h.cancel()
        signal.signal(sig, self._orig_signals[sig])
        del self._orig_signals[sig]
        return True

    # File descriptor callbacks #

    # reading from a file descriptor

    def add_reader(self, fd, callback, *args):
        """asyncio's method to call a function when a file descriptor is
        ready for reading.

        This creates a new Trio task. You may want to use "await
        :obj:`trio.lowlevel.wait_readable` (fd)" instead, or

        :param fd: Either an integer (Unix file descriptor) or an object
                   with a ``fileno`` method providing one.
        :return: A handle. To remove the listener, either call
                 ``handle.cancel()`` or :meth:`.remove_reader` (fd).
        """
        self._ensure_fd_no_transport(fd)
        return self._add_reader(fd, callback, *args)

    # Local helper to factor out common logic between _add_reader/_add_writer
    def _add_io_handler(self, set_handle, wait_ready, fd, callback, args):
        self._check_closed()
        handle = ScopedHandle(callback, args, self)
        old_handle = set_handle(fd, handle)

        if old_handle is not None:
            old_handle.cancel()
        if self._token is None:
            return None
        self._nursery.start_soon(self._io_task, fd, handle, wait_ready)
        return handle

    async def _io_task(self, fd, handle, wait_ready):
        with handle._scope:
            try:
                while True:
                    if handle._cancelled:
                        break
                    try:
                        await wait_ready(fd)
                    except OSError:
                        # maybe someone did
                        #   h = add_reader(sock); h.cancel(); sock.close()
                        # without yielding to the event loop
                        if handle._cancelled:
                            break
                        raise
                    if handle._cancelled:
                        break
                    handle._run()
                    await self.synchronize()
            except Exception as exc:
                handle._raise(exc)

    def _add_reader(self, fd, callback, *args):
        return self._add_io_handler(
            self._set_read_handle, _wait_readable, fd, callback, args
        )

    def _set_read_handle(self, fd, handle):
        try:
            key = self._selector.get_key(fd)
        except KeyError:
            self._selector.register(fd, EVENT_READ, (handle, None))
            return None
        else:
            mask, (reader, writer) = key.events, key.data
            self._selector.modify(fd, mask | EVENT_READ, (handle, writer))
            return reader

    # writing to a file descriptor

    def add_writer(self, fd, callback, *args):
        """asyncio's method to call a function when a file descriptor is
        ready for writing.

        This creates a new Trio task. You may want to use "await
        :obj:`trio.lowlevel.wait_writable` (fd) instead, or

        :param fd: Either an integer (Unix file descriptor) or an object
                   with a ``fileno`` method providing one.
        :return: A handle. To remove the listener, either call
                 ``handle.cancel()`` or :meth:`remove_writer` (fd).
        """
        self._ensure_fd_no_transport(fd)
        return self._add_writer(fd, callback, *args)

    # remove_writer: unchanged from asyncio

    def _add_writer(self, fd, callback, *args, _defer_start=False):
        return self._add_io_handler(
            self._set_write_handle, _wait_writable, fd, callback, args
        )

    def _set_write_handle(self, fd, handle):
        try:
            key = self._selector.get_key(fd)
        except KeyError:
            self._selector.register(fd, EVENT_WRITE, (None, handle))
        else:
            mask, (reader, writer) = key.events, key.data
            self._selector.modify(fd, mask | EVENT_WRITE, (reader, handle))
            return writer

    def autoclose(self, fd):
        """
        Mark a file descriptor so that it's auto-closed along with this loop.

        This is a safety measure. You also should use appropriate
        finalizers.

        Calling this method twice on the same file descriptor has no effect.

        :param fd: Either an integer (Unix file descriptor) or an object
                   with a ``fileno`` method providing one.
        """
        if hasattr(fd, "fileno"):
            fd = fd.fileno()
        self._close_files.add(fd)

    def no_autoclose(self, fd):
        """
        Un-mark a file descriptor so that it's no longer auto-closed
        along with this loop.

        Call this method either before closing the file descriptor, or when
        passing it to code out of this loop's scope.

        :param fd: Either an integer (Unix file descriptor) or an object
                   with a ``fileno()`` method providing one.
        :raises KeyError: if the descriptor is not marked to be auto-closed.
        """
        if hasattr(fd, "fileno"):
            fd = fd.fileno()
        self._close_files.remove(fd)

    # drop all file descriptors, optionally close them

    def _cancel_fds(self):
        """Clean up all of this loop's open read or write requests"""
        for fd, key in list(self._selector.get_map().items()):
            for flag in (0, 1):
                if key.events & (1 << flag):
                    handle = key.data[flag]
                    assert handle is not None
                    if not handle._cancelled:  # pragma: no branch
                        if handle._scope is not None:
                            handle._scope.cancel()
            if fd in self._close_files:
                os.close(fd)

    # The actual Trio-based main loop #

    async def _main_loop_init(self, nursery):
        """Set up the loop's internals"""
        if self._nursery is not None or not self._stopped.is_set():
            raise RuntimeError("You can't enter a loop twice")
        self._nursery = nursery
        self._task = trio.lowlevel.current_task()
        self._token = trio.lowlevel.current_trio_token()
        self._asyncgen_dispatcher = AsyncGeneratorDispatcher.install()

    async def _main_loop(self, task_status=trio.TASK_STATUS_IGNORED):
        """Run the loop by processing its event queue.

        This is the core of trio-asyncio's replacement
        of the asyncio main loop.

        Do not call this method directly.
        """

        self._stopped = trio.Event()
        task_status.started()

        try:
            # The shield here ensures that if the context surrounding
            # the loop is cancelled, we keep processing callbacks
            # until we reach the callback inserted by stop().
            # That's important to maintain the asyncio invariant
            # that everything you schedule before stop() will run
            # before the loop stops. In order to be safe against
            # deadlocks, it's important that the surrounding
            # context ensure that stop() gets called upon a
            # cancellation. (open_loop() does this indirectly
            # by calling _main_loop_exit().)
            with trio.CancelScope(shield=True):
                while not self._stopped.is_set():
                    await self._main_loop_one()
        except TrioAsyncioExit:
            # raised by .stop_me() to interrupt the loop
            pass
        finally:
            # Signal that the loop is no longer running
            self._stopped.set()

    async def _main_loop_one(self, no_wait=False):
        obj = None
        if self._timers:
            timeout = self._timers[0]._when - self.time()
            if timeout <= 0:
                # If the timer ran out, process the object now.
                obj = heapq.heappop(self._timers)
        else:
            timeout = math.inf

        if obj is None:
            if no_wait:
                obj = self._q_recv.receive_nowait()
            else:
                with trio.move_on_after(timeout):
                    obj = await self._q_recv.receive()
                if obj is None:
                    # Timeout reached. Presumably now a timer is ready,
                    # so restart from the beginning.
                    return

            if isinstance(obj, asyncio.TimerHandle):
                # A TimerHandle is added to the list of timers.
                heapq.heappush(self._timers, obj)
                return

        # assert isinstance(obj, asyncio.Handle) # speed-up
        # Hopefully one of ours
        # but it might be a standard asyncio handle

        if obj._cancelled:
            # simply skip cancelled handlers
            return

        # Don't go through the expensive nursery dance
        # if this is a sync function.
        if isinstance(obj, AsyncHandle):
            # AsyncHandle is only used to run Trio tasks, so no need to set the
            # sniffio library
            obj._context.run(self._nursery.start_soon, obj._run, name=obj._callback)
            await obj._started.wait()
        else:
            prev_library, sniffio_library.name = sniffio_library.name, "asyncio"
            try:
                obj._run()
            finally:
                sniffio_library.name = prev_library

    async def _main_loop_exit(self):
        """Finalize the loop. It may not be re-entered."""
        if self._closed:
            return

        with trio.CancelScope(shield=True):
            # wait_stopped() will return once _main_loop() exits.
            # stop() inserts a callback that will cause such, and
            # _main_loop() doesn't block except to wait for new
            # callbacks to be added, so this should be deadlock-proof.
            self.stop()
            await self.wait_stopped()

            # Drain all remaining callbacks, even those after an initial
            # call to stop(). This avoids deadlocks in some cases if
            # work is submitted to the loop after the shutdown process
            # starts. TODO: figure out precisely what this helps with,
            # maybe find a better way. test_wrong_context_manager_order
            # deadlocks if we remove it for now.
            while True:
                try:
                    await self._main_loop_one(no_wait=True)
                except trio.WouldBlock:
                    break
                except TrioAsyncioExit:
                    pass

        # Restore previous async generator hooks
        self._asyncgen_dispatcher.uninstall()
        self._asyncgen_dispatcher = None

        # Kill off unprocessed work
        self._cancel_fds()
        self._cancel_timers()

        # clean core fields
        self._nursery = None
        self._task = None
        self._token = None

    def is_running(self):
        if self._stopped is None:
            return False
        return not self._stopped.is_set()

    def run_forever(self):
        """asyncio's method to run the loop until it is stopped."""
        raise RuntimeError("This is not a sync loop")

    def run_until_complete(self, coro_or_future):
        """asyncio's method to run the loop until the coroutine returns /
        the future completes.
        """
        raise RuntimeError("This is not a sync loop")

    async def wait_stopped(self):
        """Wait until the event loop has stopped.

        This is a Trio-flavored async function. You should call it from
        somewhere outside the ``async with open_loop()`` block to avoid
        a deadlock (the event loop can't stop until all Trio tasks started
        within its scope have exited).
        """
        await self._stopped.wait()

    def _trio_io_cancel(self, cancel_scope):
        """Called when a ScopedHandle representing an I/O reader or writer
        has its cancel() method called."""
        cancel_scope.cancel()

    def stop(self):
        """Halt the main loop.

        Note that only sync loops may be restarted.
        """
        raise RuntimeError("override me")

    def close(self):
        if self._closed:
            return
        self._close()
        super().close()

    def _close(self):
        """Hook for actually closing down the concrete loop."""
        pass

    def __aenter__(self):
        raise RuntimeError("You need to use 'async with open_loop()'.")

    def __aexit__(self, *tb):
        raise RuntimeError("You need to use 'async with open_loop()'.")

    def __enter__(self):
        raise RuntimeError("You need to use a sync loop, or 'async with open_loop()'.")

    def __exit__(self, *tb):
        raise RuntimeError("You need to use a sync loop, or 'async with open_loop()'.")



import trio
import asyncio

from ._base import BaseTrioEventLoop, TrioAsyncioExit


class TrioEventLoop(BaseTrioEventLoop):
    """An asyncio event loop that runs on top of Trio, opened from
    within Trio code using :func:`open_loop`.
    """

    def _queue_handle(self, handle):
        self._check_closed()
        self._q_send.send_nowait(handle)
        return handle

    def default_exception_handler(self, context):
        """Default exception handler.

        This default handler simply re-raises an exception if there is one.

        Rationale:

        In traditional asyncio, there frequently is no context which the
        exception is supposed to affect.

        trio-asyncio, however, collects the loop and all its tasks in
        a Trio nursery. This means the context in which the error should
        be raised is known and can easily be controlled by the programmer.

        For maximum compatibility, this default handler is only used in
        asynchronous loops.

        """

        # Call the original default handler so we get the full info in the log
        super().default_exception_handler(context)

        if self._nursery is None:
            # Event loop is already closed; don't do anything further.
            # Some asyncio libraries call the asyncio exception handler
            # from their __del__ methods, e.g., aiohttp for "Unclosed
            # client session".
            return

        # Also raise an exception so it can't go unnoticed
        exception = context.get("exception")
        if exception is None:
            message = context.get("message")
            if not message:
                message = "Unhandled error in event loop"
            exception = RuntimeError(message)

        async def propagate_asyncio_error():
            raise exception

        self._nursery.start_soon(propagate_asyncio_error)

    def stop(self, waiter=None):
        """Halt the main loop.

        (Or rather, tell it to halt itself soon(ish).)

        :param waiter: an Event that is set when the loop is stopped.
        :type waiter: :class:`trio.Event`
        :return: Either the Event instance that was passed in,
                 or a newly-allocated one.

        """
        if waiter is None:
            if self._stop_wait is not None:
                return self._stop_wait
            waiter = self._stop_wait = trio.Event()
        else:
            if waiter.is_set():
                waiter.clear()

        def stop_me():
            waiter.set()
            raise TrioAsyncioExit("stopping trio-asyncio loop")

        if self._stopped.is_set():
            waiter.set()
        else:
            self._queue_handle(asyncio.Handle(stop_me, (), self))
        return waiter

    def _close(self):
        """Hook for actually closing down things."""
        if not self._stopped.is_set():
            raise RuntimeError("You need to stop the loop before closing it")
        super()._close()



# This code implements a clone of the asyncio mainloop which hooks into
# Trio.

import os
import sys
import trio
import asyncio
import warnings
import threading
from contextvars import ContextVar
from contextlib import asynccontextmanager

from ._async import TrioEventLoop
from ._util import run_aio_future

if sys.version_info < (3, 11):
    from exceptiongroup import BaseExceptionGroup

try:
    from trio.lowlevel import wait_for_child
except ImportError:
    from ._child import wait_for_child

# A substantial portion of the trio-asyncio test suite involves running the
# stock asyncio test suite with trio-asyncio imported. This is intended to
# test two things:
# - that trio-asyncio is a "good citizen": won't screw up other users of
#   asyncio if Trio isn't running
# - that trio-asyncio provides an event loop that conforms to asyncio semantics,
#   even if Trio is running
#
# It's hard to test both of these at once: in order to get good test
# coverage, we want normal asyncio calls to instantiate our loop, but
# the asyncio tests are full of tricky event loop manipulations, some
# of which expect to provide their own mock loop. We've compromised on
# the following.
#
# - The "actual" event loop policy (the one that would be returned by
#   an unpatched asyncio.get_event_loop_policy()) is set at import
#   time to a singleton instance of TrioPolicy, and never changed
#   later. This is required for correct operation in CPython 3.7+,
#   because the C _asyncio module caches the original
#   asyncio.get_event_loop_policy() and calls it from its accelerated
#   C get_event_loop() function. We want asyncio.get_event_loop() to be
#   able to return a trio-asyncio event loop.
#
# - To cope with tests that set a custom policy, we monkeypatch
#   asyncio.get_event_loop_policy() and set_event_loop_policy()
#   so that they model an event loop policy that is thread-local when
#   called outside of Trio context. Said policy is stored in at
#   _faked_policy.policy. (Inside Trio context, we let get_event_loop_policy()
#   return the singleton global TrioPolicy, and set_event_loop_policy()
#   raises an exception.)
#
#   - If you've previously called set_event_loop_policy() with a
#     non-None argument in the current thread, then
#     get_event_loop_policy() will return the thing that you passed to
#     set_event_loop_policy().
#
#   - If you haven't called set_event_loop_policy() in this thread
#     yet, or the most recent call in this context had a None
#     argument, then get_event_loop_policy() will return the asyncio
#     event loop policy that was installed when trio_asyncio was
#     imported.
#
# - Even though the user can set a per-thread policy and we'll echo it back,
#   the "actual" global policy is still the TrioPolicy and we don't expose
#   any way to change it. asyncio.get_event_loop() will use this TrioPolicy
#   on 3.7+ no matter what we do, so we monkeypatch new_event_loop() and
#   set_event_loop() to go through the TrioPolicy too (for consistency's sake)
#   and let TrioPolicy forward to the appropriate actual policy specified by
#   the user.
#
#   - Inside a Trio context, TrioPolicy refuses to create a new event loop
#     (you should use 'async with trio_asyncio.open_loop():' instead).
#     Its get/set event loop methods access a contextvar (current_loop),
#     which is normally set to the nearest enclosing open_loop() loop,
#     but can be modified if you want to put some Trio tasks in a
#     trio-asyncio event loop that doesn't correspond to their place in the
#     Trio task tree.
#
#   - Outside a Trio context when an event loop policy has been set,
#     TrioPolicy delegates all three methods (new/get/set event loop)
#     to that policy. Thus, if you install a custom policy, it will get
#     used (trio-asyncio gets out of the way).
#
#   - Outside a Trio context when no event loop policy has been set,
#     the get/set event loop methods manage a thread-local event loop
#     just like they do in default asyncio. However, new_event_loop() will
#     create a synchronous trio-asyncio event loop (the kind that can
#     be repeatedly started and stopped, which is helpful for many asyncio
#     tests). Thus, if you don't install a custom policy, tests that use
#     asyncio will exercise trio-asyncio.


class _FakedPolicy(threading.local):
    policy = None


_faked_policy = _FakedPolicy()


def _in_trio_context():
    try:
        trio.lowlevel.current_task()
    except RuntimeError:
        return False
    else:
        return True


_sync_loop_task_name = "trio_asyncio sync loop task"


def _in_trio_context_other_than_sync_loop():
    try:
        return trio.lowlevel.current_task().name != _sync_loop_task_name
    except RuntimeError:
        return False


class _TrioPolicy(asyncio.events.BaseDefaultEventLoopPolicy):
    @staticmethod
    def _loop_factory():
        raise RuntimeError("Event loop creations shouldn't get here")

    def new_event_loop(self):
        if _in_trio_context_other_than_sync_loop():
            raise RuntimeError(
                "You're within a Trio environment.\n"
                "Use 'async with open_loop()' instead."
            )
        if _faked_policy.policy is not None:
            return _faked_policy.policy.new_event_loop()

        from ._sync import SyncTrioEventLoop

        return SyncTrioEventLoop()

    def get_event_loop(self):
        """Get the current event loop.

        Note that this will auto-generate an event loop if none exists, for
        compatibility with asyncio.

        To get a Trio-compatible asyncio loop, use
        ``async with trio_asyncio.open_loop() as loop:``.

        To test whether an event loop is running, check the loop policy's
        ``.current_event_loop`` property.
        """
        try:
            task = trio.lowlevel.current_task()
        except RuntimeError:
            pass
        else:
            # Trio context. Note: NOT current_loop.get()! If this is called from
            # asyncio code, current_task() is the trio-asyncio loop runner task,
            # which has the correct loop set in its contextvar; but (on Python
            # 3.7+) our current context is quite possibly something different, and
            # might have the wrong contextvar value (e.g. in the case of a
            # loop1.call_later() in loop2's context).
            return task.context.get(current_loop)

        # Not Trio context
        if _faked_policy.policy is not None:
            return _faked_policy.policy.get_event_loop()

        # This will return the thread-specific event loop set using
        # set_event_loop(), or if none has been set, will call back into
        # our new_event_loop() to make a SyncTrioEventLoop and set it as
        # this thread's event loop.
        return super().get_event_loop()

    @property
    def current_event_loop(self):
        """The currently-running event loop, if one exists."""
        loop = current_loop.get()
        if loop is None:
            loop = super().get_event_loop()
        return loop

    def set_event_loop(self, loop):
        """Set the current event loop."""
        if _in_trio_context():
            current_loop.set(loop)
        elif _faked_policy.policy is not None:
            _faked_policy.policy.set_event_loop(loop)
        else:
            super().set_event_loop(loop)


# get_event_loop() without a running loop is deprecated in 3.12+. The logic for emitting the
# DeprecationWarning walks the stack looking at module names in order to associate it with
# the first caller outside asyncio. We need to pretend to be asyncio in order for that to work.
if sys.version_info >= (3, 12):
    __name__ = "asyncio.fake.trio_asyncio._loop"

# Make sure we don't try to continue using the Trio loop after a fork()
if hasattr(os, "register_at_fork"):

    def _clear_state_after_fork():
        if _in_trio_context():
            from trio._core._run import GLOBAL_RUN_CONTEXT

            del GLOBAL_RUN_CONTEXT.task
            del GLOBAL_RUN_CONTEXT.runner
            current_loop.set(None)

    os.register_at_fork(after_in_child=_clear_state_after_fork)

from asyncio import events as _aio_event

#####


def _new_policy_get():
    if _in_trio_context():
        return _trio_policy
    elif _faked_policy.policy is not None:
        return _faked_policy.policy
    else:
        return _original_policy


def _new_policy_set(new_policy):
    if isinstance(new_policy, TrioPolicy):
        raise RuntimeError("You can't set the Trio loop policy manually")
    if _in_trio_context_other_than_sync_loop():
        raise RuntimeError("You can't change the event loop policy in Trio context")
    if new_policy is not None and not isinstance(
        new_policy, asyncio.AbstractEventLoopPolicy
    ):
        # Raise the type of error that the CPython test suite expects
        raise_type = TypeError if sys.version_info >= (3, 11) else AssertionError
        raise raise_type(
            "policy must be an instance of AbstractEventLoopPolicy or None, "
            f"not '{type(new_policy).__name__}'"
        )
    _faked_policy.policy = new_policy


_orig_policy_get = _aio_event.get_event_loop_policy
_orig_policy_set = _aio_event.set_event_loop_policy
_aio_event.get_event_loop_policy = _new_policy_get
_aio_event.set_event_loop_policy = _new_policy_set
asyncio.get_event_loop_policy = _new_policy_get
asyncio.set_event_loop_policy = _new_policy_set

#####

_orig_run_get = _aio_event._get_running_loop


def _new_run_get():
    try:
        task = trio.lowlevel.current_task()
    except RuntimeError:
        pass
    else:
        # Trio context. Note: NOT current_loop.get()!
        # See comment in _TrioPolicy.get_event_loop().
        return task.context.get(current_loop)
    # Not Trio context
    return _orig_run_get()


# Must override the non-underscore-prefixed get_running_loop() too,
# else will use the C-accelerated one which doesn't call the patched
# _get_running_loop()
def _new_run_get_or_throw():
    result = _new_run_get()
    if result is None:
        raise RuntimeError("no running event loop")
    return result


_aio_event._get_running_loop = _new_run_get
_aio_event.get_running_loop = _new_run_get_or_throw
asyncio._get_running_loop = _new_run_get
asyncio.get_running_loop = _new_run_get_or_throw

#####


def _new_loop_get():
    current_loop = _new_run_get()
    if current_loop is not None:
        return current_loop
    return _trio_policy.get_event_loop()


def _new_loop_set(new_loop):
    _trio_policy.set_event_loop(new_loop)


def _new_loop_new():
    return _trio_policy.new_event_loop()


_orig_loop_get = _aio_event.get_event_loop
_orig_loop_set = _aio_event.set_event_loop
_orig_loop_new = _aio_event.new_event_loop
_aio_event.get_event_loop = _new_loop_get
_aio_event.set_event_loop = _new_loop_set
_aio_event.new_event_loop = _new_loop_new
asyncio.get_event_loop = _new_loop_get
asyncio.set_event_loop = _new_loop_set
asyncio.new_event_loop = _new_loop_new

# current_task is implemented in C in 3.12+, which creates a problem because it
# accesses the non-monkeypatched version of _get_running_loop()
from asyncio import current_task as _orig_current_task


def _new_current_task(loop=None):
    return _orig_current_task(loop or _new_run_get())


asyncio.tasks.current_task = _new_current_task
asyncio.current_task = _new_current_task

#####


class TrioPolicy(_TrioPolicy, asyncio.DefaultEventLoopPolicy):
    """This is the loop policy that's active whenever we're in a Trio context."""

    def _init_watcher(self):
        with asyncio.events._lock:
            if self._watcher is None:  # pragma: no branch
                self._watcher = TrioChildWatcher()
                if isinstance(threading.current_thread(), threading._MainThread):
                    self._watcher.attach_loop(current_loop.get())

        if self._watcher is not None and isinstance(
            threading.current_thread(), threading._MainThread
        ):
            self._watcher.attach_loop(current_loop.get())

    def set_child_watcher(self, watcher):
        if watcher is not None:
            if not isinstance(watcher, TrioChildWatcher):
                # raise RuntimeError("You must use a TrioChildWatcher here. "
                #                    "Sorry.")
                # warnings.warn("You must use a TrioChildWatcher.")
                #
                loop = watcher._loop  # ugh.
                watcher.close()
                watcher = TrioChildWatcher()
                watcher.attach_loop(loop)
        super().set_child_watcher(watcher)


_original_policy = _orig_policy_get()
_trio_policy = TrioPolicy()
_orig_policy_set(_trio_policy)

# Backwards compatibility -- unused
current_policy = ContextVar("trio_aio_policy", default=_trio_policy)

current_loop = ContextVar("trio_aio_loop", default=None)


class TrioChildWatcher(
    asyncio.AbstractChildWatcher if sys.platform != "win32" else object
):
    """Watches for child processes to exit using Trio APIs.

    All TrioChildWatchers behave identically, so there's no reason to construct
    your own. This is more or less an implementation detail, exposed publicly
    because you can get your hands on it anyway (using
    ``asyncio.get_event_loop_policy
    """

    # AbstractChildWatcher not available under Windows
    def __init__(self):
        super().__init__()
        self._callbacks = {}  # pid => handler

    def attach_loop(self, loop):
        self._loop = loop

    async def _waitpid(self, pid, callback, *args):
        returncode = await wait_for_child(pid)
        callback(pid, returncode, *args)

    def add_child_handler(self, pid, callback, *args):
        """Add a callback to run when a child process terminates."""
        h = self._loop.trio_as_future(self._waitpid, pid, callback, *args)
        self._callbacks[pid] = h

    def remove_child_handler(self, pid):
        """Remove the callback to run when a child process terminates."""
        h = self._callbacks.pop(pid, None)
        if h is None:
            return False
        h.cancel()
        return True

    def close(self):
        for pid in list(self._callbacks):
            h = self._callbacks.pop(pid, None)
            if h is None:
                continue
            h.cancel()
        self._loop = None

    def __enter__(self):
        return self

    def __exit__(self, *tb):
        self.close()


@asynccontextmanager
async def open_loop(queue_len=None):
    """Returns a Trio-flavored async context manager which provides
    an asyncio event loop running on top of Trio.

    The context manager evaluates to a new `TrioEventLoop` object.

    Entering the context manager is not enough on its own to immediately
    run asyncio code; it just provides the context that makes running that
    code possible. You additionally need to wrap any asyncio functions
    that you want to run in :func:`aio_as_trio`.

    If you provide a *queue_len*, then any attempt to enqueue more
    than that many asyncio callbacks near-simultaneously (including,
    for example, new task creations) will fail. There is no way to
    backpressure asyncio callback registration, so the best we can
    do if the queue length is exceeded is raise an exception
    (`trio.WouldBlock`), which is likely to crash your whole program.
    It is suggested to leave the *queue_len* at its default of None
    (unlimited) unless you need to enforce hard constraints on memory use.

    Exiting the context manager will attempt to do an orderly shutdown
    of the tasks it contains, analogously to :func:`asyncio.run`.
    Both asyncio-flavored tasks and Trio-flavored tasks (the latter
    started using :meth:`~BaseTrioEventLoop.trio_as_future`,
    :meth:`~BaseTrioEventLoop.run_trio_task`, :func:`trio_as_aio`,
    etc) are cancelled simultaneously, and the loop waits for them to
    exit in response to this cancellation before proceeding. All
    :meth:`~asyncio.loop.call_soon` callbacks that are submitted
    before exiting the context manager will run before starting this
    shutdown sequence, and all callbacks that are submitted before the
    last task exits will run before the loop closes.  The exact point
    at which the loop stops running callbacks is not specified.

    .. warning:: As with :func:`asyncio.run`, asyncio-flavored tasks
       that are started *after* exiting the context manager (such as by
       another task as it unwinds) may or may not be cancelled, and will
       be abandoned if they survive the shutdown sequence. This may lead
       to unclosed resources, stderr spew about "coroutine ignored
       GeneratorExit", etc. Trio-flavored tasks do not have this hazard.

    Example usage::

            async def async_main(*args):
                async with trio_asyncio.open_loop() as loop:
                    # async part of your main program here
                    await trio.sleep(1)
                    await trio_asyncio.aio_as_trio(asyncio.sleep)(2)

    """

    # TODO: make sure that there is no asyncio loop already running

    # The trio-asyncio loop can't shut down until all trio_as_aio tasks
    # (or others using run_trio) have exited. This is because the
    # termination of such a Trio task sets an asyncio future, which
    # uses call_soon(), which won't work if the loop is closed.
    # So, we use two nested nurseries.
    async with trio.open_nursery() as loop_nursery:
        loop = TrioEventLoop(queue_len=queue_len)
        old_loop = current_loop.set(loop)
        forwarded_cancellation = None
        try:
            loop._closed = False
            async with trio.open_nursery() as tasks_nursery:
                # Shield Trio background tasks and I/O waiters from
                # immediate cancellation if the entire open_loop is
                # cancelled. They should be cancelled based on the
                # cancellation status of the asyncio task that ran them.
                tasks_nursery.cancel_scope.shield = True

                # Note that this shield also has the side effect of guaranteeing
                # that the start() call below actually moves the _main_loop
                # task into the tasks_nursery. TaskStatus.started()
                # doesn't complete Nursery.start() if there's a
                # cancellation pending, because it figures the task
                # will be cancelled soon enough and doesn't want to
                # worry about Cancelled exceptions propagating to the
                # wrong place; but _main_loop shields everything it does
                # after started(), so this just results in start() never
                # completing. With the shield here, started() can't see
                # the outer cancellation, which avoids the deadlock.
                await loop._main_loop_init(tasks_nursery)
                await loop_nursery.start(loop._main_loop)

                try:
                    # Since we're inside the tasks_nursery which is shielded,
                    # we need to forward cancellation from outside open_loop()
                    # into its body.
                    forward_cancel_scope = trio.CancelScope()
                    with trio.CancelScope() as body_scope:

                        @loop_nursery.start_soon
                        async def forward_cancellation():
                            try:
                                with forward_cancel_scope:
                                    await trio.sleep_forever()
                            except trio.Cancelled:
                                # We only want a trio.Cancelled to escape from
                                # open_loop() if something inside the body took
                                # a cancellation
                                pass
                            finally:
                                body_scope.cancel()

                        try:
                            yield loop
                        except trio.Cancelled as exc:
                            forwarded_cancellation = exc
                        except BaseExceptionGroup as exc:
                            forwarded_cancellation, rest = exc.split(trio.Cancelled)
                            if rest is not None:
                                raise rest
                        finally:
                            forward_cancel_scope.cancel()

                finally:
                    # Allow all already-submitted tasks a chance to start
                    # (and then immediately be cancelled), unless the loop
                    # stops (due to someone else calling stop()) before
                    # that.
                    async with trio.open_nursery() as sync_nursery:

                        @sync_nursery.start_soon
                        async def wait_for_sync():
                            if not loop.is_closed():
                                await loop.synchronize()
                            sync_nursery.cancel_scope.cancel()

                        await loop.wait_stopped()
                        sync_nursery.cancel_scope.cancel()

                    # Cancel and wait on all currently-running tasks.
                    # Exiting the tasks_nursery will wait for the Trio tasks
                    # automatically; we mix in the asyncio tasks by scheduling
                    # a call to run_aio_future() for each one. It's important
                    # not to wait on one kind of task before the other, so that
                    # we support Trio tasks that need to run some asyncio
                    # code during teardown as well as the opposite.
                    # Like asyncio.run(), we don't bother cancelling and waiting
                    # on any additional asyncio tasks that these tasks start
                    # as they unwind.
                    aio_tasks = asyncio.all_tasks(loop)
                    for task in aio_tasks:
                        tasks_nursery.start_soon(run_aio_future, task)
                    tasks_nursery.cancel_scope.cancel()

        finally:
            # If we have any async generators left, finalize them before
            # closing the event loop. Make sure that the finalizers have a
            # chance to actually start before they're exposed to any
            # external cancellation, since asyncio doesn't guarantee that
            # cancelled tasks have a chance to start first.

            asyncgens_done = trio.Event()
            should_warn = False
            if len(loop._asyncgens) == 0:
                asyncgens_done.set()
            elif not loop.is_running():
                asyncgens_done.set()
                should_warn = True
            else:
                shield_asyncgen_finalizers = trio.CancelScope(shield=True)

                async def sentinel():
                    try:
                        yield
                    finally:
                        try:
                            # Open-coded asyncio version of loop.synchronize();
                            # since we closed the tasks_nursery, we can't do
                            # any more asyncio-to-trio-mode conversions
                            w = asyncio.Event()
                            loop.call_soon(w.set)
                            await w.wait()
                        finally:
                            shield_asyncgen_finalizers.shield = False

                async def shutdown_asyncgens_from_aio():
                    agen = sentinel()
                    await agen.asend(None)
                    try:
                        await loop.shutdown_asyncgens()
                    finally:
                        asyncgens_done.set()

                @loop_nursery.start_soon
                async def shutdown_asyncgens_from_trio():
                    with shield_asyncgen_finalizers:
                        await loop.run_aio_coroutine(shutdown_asyncgens_from_aio())

            if forwarded_cancellation is not None:
                # Now that we're outside the shielded tasks_nursery, we can
                # add this cancellation to the set of errors propagating out
                # of the loop_nursery.

                @loop_nursery.start_soon
                async def forward_cancellation():
                    raise forwarded_cancellation

            try:
                try:
                    if should_warn:
                        warnings.warn(
                            "trio-asyncio loop was stopped before its async "
                            "generators were finalized; weird stuff might happen",
                            RuntimeWarning,
                        )
                finally:
                    with trio.CancelScope(shield=True):
                        await asyncgens_done.wait()
                    await loop._main_loop_exit()
            finally:
                loop.close()
                current_loop.reset(old_loop)


def run(proc, *args, queue_len=None, **trio_run_options):
    """Run a Trio-flavored async function in a context that has an
    asyncio event loop also available.

    This is exactly equivalent to using :func:`trio.run` plus wrapping
    the body of *proc* in ``async with trio_asyncio.open_loop():``.
    See the documentation of :func:`open_loop` for more on the *queue_len*
    argument, which should usually be left at its default of None.
    """

    async def _run_task(proc, args):
        async with open_loop(queue_len=queue_len):
            return await proc(*args)

    return trio.run(_run_task, proc, args, **trio_run_options)


# Non-deprecated aliases for event loop methods


def _running_loop():
    loop = current_loop.get()
    if loop is None:
        raise RuntimeError("You are not within a trio_asyncio loop")
    return loop


async def run_aio_coroutine(coro):
    """Alias for a call to :meth:`~BaseTrioEventLoop.run_aio_coroutine`
    on the event loop returned by :func:`asyncio.get_event_loop`.

    This is a Trio-flavored async function which takes an asyncio-flavored
    coroutine object.
    """
    return await _running_loop().run_aio_coroutine(coro)


def run_trio(proc, *args):
    """Alias for a call to :meth:`~BaseTrioEventLoop.trio_as_future`
    on the event loop returned by :func:`asyncio.get_event_loop`.

    This is a synchronous function which takes a Trio-flavored async function
    and returns an asyncio Future.
    """
    return _running_loop().trio_as_future(proc, *args)


def run_trio_task(proc, *args):
    """Alias for a call to :meth:`~BaseTrioEventLoop.run_trio_task`
    on the event loop returned by :func:`asyncio.get_event_loop`.

    This is a synchronous function which takes a Trio-flavored async
    function and returns nothing (the handle returned by
    `BaseTrioEventLoop.run_trio_task` is discarded). An uncaught error
    will propagate to, and terminate, the trio-asyncio loop.
    """
    _running_loop().run_trio_task(proc, *args)



import sys
import trio
import types
import asyncio
from asyncio.format_helpers import _format_callback, _get_function_source

if sys.version_info < (3, 11):
    from exceptiongroup import BaseExceptionGroup


def _format_callback_source(func, args):
    func_repr = _format_callback(func, args, None)
    source = _get_function_source(func)
    if source:  # pragma: no cover
        func_repr += " at %s:%s" % source
    return func_repr


class ScopedHandle(asyncio.Handle):
    """An asyncio.Handle that cancels a trio.CancelScope when the Handle is cancelled.

    This is used to manage installed readers and writers, so that the Trio call to
    wait_readable() or wait_writable() can be cancelled when the handle is.
    """

    __slots__ = ("_scope",)

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self._scope = trio.CancelScope()

    def cancel(self):
        super().cancel()
        self._loop._trio_io_cancel(self._scope)

    def _repr_info(self):
        return super()._repr_info() + ["scope={!r}".format(self._scope)]

    def _raise(self, exc):
        """This is a copy of the exception handling in asyncio.events.Handle._run().
        It's used to report exceptions that arise when waiting for readability
        or writability, and exceptions in async tasks managed by our subclass
        AsyncHandle.
        """
        cb = _format_callback_source(self._callback, self._args)
        msg = "Exception in callback {}".format(cb)
        context = {
            "message": msg,
            "exception": exc,
            "handle": self,
        }
        if self._source_traceback:
            context["source_traceback"] = self._source_traceback
        self._loop.call_exception_handler(context)


# copied from trio._core._multierror, but relying on traceback constructability
# from Python (as introduced in 3.7) instead of ctypes hackery
def concat_tb(head, tail):
    # We have to use an iterative algorithm here, because in the worst case
    # this might be a RecursionError stack that is by definition too deep to
    # process by recursion!
    head_tbs = []
    pointer = head
    while pointer is not None:
        head_tbs.append(pointer)
        pointer = pointer.tb_next
    current_head = tail
    for head_tb in reversed(head_tbs):
        current_head = types.TracebackType(
            current_head, head_tb.tb_frame, head_tb.tb_lasti, head_tb.tb_lineno
        )
    return current_head


# copied from trio._core._run with minor modifications:
def collapse_exception_group(excgroup):
    """Recursively collapse any single-exception groups into that single contained
    exception.
    """
    exceptions = list(excgroup.exceptions)
    modified = False
    for i, exc in enumerate(exceptions):
        if isinstance(exc, BaseExceptionGroup):
            new_exc = collapse_exception_group(exc)
            if new_exc is not exc:
                modified = True
                exceptions[i] = new_exc

    collapse = getattr(excgroup, "collapse", False) or (  # Trio 0.23.0  # Trio 0.24.0
        getattr(trio._core._run, "NONSTRICT_EXCEPTIONGROUP_NOTE", object())
        in getattr(excgroup, "__notes__", ())
    )
    if len(exceptions) == 1 and collapse:
        exceptions[0].__traceback__ = concat_tb(
            excgroup.__traceback__, exceptions[0].__traceback__
        )
        return exceptions[0]
    elif modified:
        return excgroup.derive(exceptions)
    else:
        return excgroup


def collapse_aware_exception_split(exc, etype):
    if not isinstance(exc, BaseExceptionGroup):
        if isinstance(exc, etype):
            return exc, None
        else:
            return None, exc

    match, rest = exc.split(etype)
    if isinstance(match, BaseExceptionGroup):
        match = collapse_exception_group(match)
    if isinstance(rest, BaseExceptionGroup):
        rest = collapse_exception_group(rest)
    return match, rest


class AsyncHandle(ScopedHandle):
    """A ScopedHandle associated with the execution of a Trio-flavored
    async function.

    If the handle is cancelled, the cancel scope surrounding the async function
    will be cancelled too. It is also possible to link a future to the result
    of the async function. If you do that, the future will evaluate to the
    result of the function, and cancelling the future will cancel the handle too.

    """

    __slots__ = ("_fut", "_started", "_finished", "_cancel_message")

    def __init__(self, *args, result_future=None, **kw):
        super().__init__(*args, **kw)
        self._fut = result_future
        self._started = trio.Event()
        self._finished = False
        self._cancel_message = None

        if self._fut is not None:
            orig_cancel = self._fut.cancel

            def wrapped_cancel(msg=None):
                if self._finished:
                    # We're being called back after the task completed
                    if msg is not None:
                        return orig_cancel(msg)
                    elif self._cancel_message is not None:
                        return orig_cancel(self._cancel_message)
                    else:
                        return orig_cancel()
                if self._fut.done():
                    return False
                # Forward cancellation to the Trio task, don't mark
                # future as cancelled until it completes
                self._cancel_message = msg
                self.cancel()
                return True

            self._fut.cancel = wrapped_cancel

    async def _run(self):
        self._started.set()
        if self._cancelled:
            self._finished = True
            return

        try:
            # Run the callback
            with self._scope:
                try:
                    res = await self._callback(*self._args)
                finally:
                    self._finished = True

            if self._fut:
                # Propagate result or just-this-handle cancellation to the Future
                if self._scope.cancelled_caught:
                    self._fut.cancel()
                elif not self._fut.cancelled():
                    self._fut.set_result(res)

        except BaseException as exc:
            if not self._fut:
                # Pass Exceptions through the fallback exception
                # handler since they have nowhere better to go. (In an
                # async loop this will still raise the exception out
                # of the loop, terminating it.) Let BaseExceptions
                # escape so that Cancelled and SystemExit work
                # reasonably.
                rest, base = collapse_aware_exception_split(exc, Exception)
                if rest:
                    self._raise(rest)
                if base:
                    raise base
            else:
                # The result future gets all the non-Cancelled
                # exceptions.  Any Cancelled need to keep propagating
                # out of this stack frame in order to reach the cancel
                # scope for which they're intended.  Any non-Cancelled
                # BaseExceptions keep propagating.
                cancelled, rest = collapse_aware_exception_split(exc, trio.Cancelled)
                if not self._fut.cancelled():
                    if rest:
                        self._fut.set_exception(rest)
                    else:
                        self._fut.cancel()
                if cancelled:
                    raise cancelled

        finally:
            # asyncio says this is needed to break cycles when an exception occurs.
            # I'm not so sure, but it doesn't seem to do any harm.
            self = None
